# This software is a part of NOODLYBOX.
# This software is distributed under the terms of the new BSD License.
# Copyright (c) 2009, 2011, 2012 molelord
# All rights reserved.

# echo canceller
proc silent args {
}

silent [set nbox_interpreter ""]
proc detect_interpreter {} {
    global nbox_interpreter
    if {$nbox_interpreter ne ""} {
        return $nbox_interpreter
    } elseif {[info commands wlfman] eq "wlfman"} {
        set nbox_interpreter "msim"
    } elseif {[info commands virtualbus] eq "virtualbus"} {
        set nbox_interpreter "isim"
    } elseif {[info commands asdb2awf] eq "asdb2awf"} {
        set nbox_interpreter "ahdl"
    }
    return $nbox_interpreter
}

# Put the debug information.
proc putdbg {str} {
    if {$::dbg} { puts stderr $str }
}

proc isSystemVerilog {filename} {
    # Search "DPI-C"
    set iChannel [open $filename]
    set found 0
    while {1} {
        set rc [gets $iChannel line]
        if {$rc == -1} {
            break
        }
        # Ignore any comments
        regsub "//.*" $line "" line

        if {[regexp "\"DPI-C\"" $line]} {
            set found 1
        }
    }
    close $iChannel
    return $found
}

proc adjustgrid_by {filename} {
    set iChannel [open $filename]
    set match  0
    set period 0
    while {1} {
        set rc [gets $iChannel line]
        if {$rc == -1} {
            break
        }
        # Ignore any comments
        regsub "//.*" $line "" line

        # Searching clock period
        # for Verilog HDL
        if {[regexp "localparam.*STEP.*?(\\d+)" $line ignore match]} {
            putdbg "-- gridperiod : $match --"
        }
        # for VHDL
        if {[regexp "constant.*STEP.*?(\\d+)" $line ignore match]} {
            putdbg "-- gridperiod : $match --"
        }
    }
    close $iChannel

    if {$match} {
        set full $match
        set half [expr $full / 2]
        append full "ns"
        append half "ns"
        configure wave -gridoffset $half -gridperiod $full -timeline 1
    }
}

proc get_result_from_log {} {
    set result {}
    if {[detect_interpreter] eq "isim"} {
        # The subsequent lines of last '#' will be reflected in the result.
        set exp      {^# }
        set iChannel [open "isim.log" "r"]
        while {[gets $iChannel line] != -1} {
            if {[regexp -- $exp $line]} {
                set result {}
            } else {
                lappend result $line
            }
        }
        close $iChannel
    } elseif {[detect_interpreter] eq "ahdl"} {
        global dsn
        set exp1     {^# gather_wave}
        set exp2     {;\s*gather_wave}
        flushlogs
        set iChannel [open "$dsn/log/console.log" "r"]
        while {[gets $iChannel line] != -1} {
            if {[regexp -- $exp1 $line] || [regexp -- $exp2 $line]} {
                set result {}
            } else {
                lappend result $line
            }
        }
        close $iChannel
    }
    return $result
}

proc gather_wave_isim {args} {
    show child -r

    set result [get_result_from_log]
    set hier [list]
    set ignore_level -1

    foreach i $result {
        regexp "^( +)(?:VHDL|Verilog)" $i dummy level
        set level [expr [string length $level] / 4 - 1]

        if {$ignore_level != -1 && $level > $ignore_level} {
            continue
        } else {
            set ignore_level -1
            set internal 0
            set instance 0

            if {[regexp "^ +(?:VHDL|Verilog) Internal Block: \{(.+)\}" $i dummy match]} {
                set internal 1
            } elseif {[regexp "^ +(?:VHDL|Verilog) Instance: \{(.+)\}" $i dummy match]} {
                set instance 1
            }
            if {$internal || $instance} {
                set len [llength $hier]
                if {$level + 1 <= $len} {
                    set hier [lreplace $hier $level $level $match]
                } else {
                    lappend hier "$match"
                }
                #puts stdout $level

                set path ""
                foreach j [lrange $hier 0 $level] {
                    append path "/$j"
                }

                set bboxlist [split $args "\{\} "]
                foreach j $bboxlist {
                    #puts stdout "<$j>"
                    if {$j ne "" && [regexp -nocase -- $j $path]} {
                        set ignore_level $level
                        break
                    }
                }

                if {$ignore_level == -1 && $instance} {
                    set gid [group add $path]
                    if { [catch {
                        wave add $path -radix hex -into $gid
                    } fid] } {
                        # do nothing
                    }
                }
            }
        }
    }
}

proc gather_wave_msim {args} {
    set result [Show -all]
    set str_num 0
    foreach i $result {
        #puts stdout $i
        incr str_num
        if {$str_num == 1} {
            continue
        } elseif {$str_num == 2} {
            set hierarchy $i
            add wave -hex -group $hierarchy $hierarchy/*
            continue
        }

        set start 0
        set level 0
        # Search { or }
        while {[regexp -start $start -indices {[\{\}]} $i range]} {
            set pos [lindex $range 0]
            set ch [string index $i $pos]
            set start [expr $pos + 1]

            if {$ch eq "\{"} {
                incr level
                incr pos
                if {([string first "Module"          $i $pos] == $pos) ||
                    ([string first "VlGenerateBlock" $i $pos] == $pos) ||
                    ([string first "Statement"       $i $pos] == $pos)} {
                    regexp -start $pos { (\S+)} $i ignore name
                    set hierarchy $hierarchy/$name
                    #puts stdout "$level Module:$hierarchy"

                    set skip 0
                    set bboxlist [split $args "\{\} "]
                    foreach j $bboxlist {
                        if {$j ne "" && [regexp -nocase -- $j $hierarchy]} {
                            set skip 1
                            break
                        }
                    }
                    if {$skip} continue

                    if {[catch {
                        add wave -hex -group $hierarchy $hierarchy/*
                    } fid]} {
                        # The display fails when no signals are included.
                        delete wave $hierarchy
                    }
                }
            } else {
                set level [expr $level - 1]
                if {[expr $level % 2] == 1} {
                    set hierarchy [regsub {/[^/]+$} $hierarchy ""]
                }
            }
        }
    }
}

proc gather_wave_ahdl {args} {
    show -rec
    set result [get_result_from_log]
    set exp {(\S+) : (\S+) \((\S+)\)}
    #puts stdout $result
    foreach i $result {
        #puts stdout $i
        if {[regexp $exp $i ignore hierarchy name1 name2]} {
            #puts stdout "h: $hierarchy n1: $name1 n2: $name2"

            set skip 0
            set bboxlist [split $args "\{\} "]
            foreach j $bboxlist {
                if {$j ne "" && [regexp -nocase -- $j $hierarchy]} {
                    set skip 1
                    break
                }
            }
            if {$skip} continue

            if {$hierarchy eq "Root"} {
                wave -hex -vgroup / /*
            } else  {
                wave -hex -vgroup $hierarchy $hierarchy/*
            }
        }
    }
}

# Usage : gather_wave [blackbox_instance1 blackbox_instance2 ...]
proc gather_wave {args} {
    set suffix [detect_interpreter]
    if {$suffix ne ""} {
        gather_wave_$suffix $args
    }
}

proc dump_pblaz {instance_name} {
    divider add "PicoBlaze"

    wave add $instance_name/pc -radix hex
    wave add $instance_name/instruction -radix hex
    wave add $instance_name/port_id -radix hex
    wave add $instance_name/write_strobe
    wave add $instance_name/out_port -radix hex
    wave add $instance_name/read_strobe
    wave add $instance_name/in_port -radix hex
    wave add $instance_name/interrupt
    wave add $instance_name/interrupt_ack

    # Dump s0-sF registers
    for {set reg 0} {$reg < 16} {incr reg} {
        set name [format "s%X" $reg]
        set vbusid [virtualbus add $name -radix hex]
        wave add -into $vbusid \
        $instance_name/\\reg_loop(7)\\/register_bit/mem\[$reg\] \
        $instance_name/\\reg_loop(6)\\/register_bit/mem\[$reg\] \
        $instance_name/\\reg_loop(5)\\/register_bit/mem\[$reg\] \
        $instance_name/\\reg_loop(4)\\/register_bit/mem\[$reg\] \
        $instance_name/\\reg_loop(3)\\/register_bit/mem\[$reg\] \
        $instance_name/\\reg_loop(2)\\/register_bit/mem\[$reg\] \
        $instance_name/\\reg_loop(1)\\/register_bit/mem\[$reg\] \
        $instance_name/\\reg_loop(0)\\/register_bit/mem\[$reg\] \
    }

    wave add $instance_name/zero_flag
    wave add $instance_name/carry_flag

    # Dump the scratch pad
    for {set sc 0} {$sc < 64} {incr sc} {
        set name [format "scratchpad%02X" $sc]
        set vbusid [virtualbus add $name -radix hex]
        wave add -into $vbusid \
        $instance_name/\\store_loop(7)\\/memory_bit/mem\[$sc\] \
        $instance_name/\\store_loop(6)\\/memory_bit/mem\[$sc\] \
        $instance_name/\\store_loop(5)\\/memory_bit/mem\[$sc\] \
        $instance_name/\\store_loop(4)\\/memory_bit/mem\[$sc\] \
        $instance_name/\\store_loop(3)\\/memory_bit/mem\[$sc\] \
        $instance_name/\\store_loop(2)\\/memory_bit/mem\[$sc\] \
        $instance_name/\\store_loop(1)\\/memory_bit/mem\[$sc\] \
        $instance_name/\\store_loop(0)\\/memory_bit/mem\[$sc\]
    }
}
